@@ -1,4 +1,4 @@
-/* $Id: remap.c,v 1.11 2002/03/08 04:08:05 hpa Exp $ */
+/* $Id: remap.c,v 1.12 2002/05/06 19:14:14 hpa Exp $ */
 /* ----------------------------------------------------------------------- *
  *   
  *   Copyright 2001 H. Peter Anvin - All Rights Reserved
@@ -146,7 +146,7 @@ static int parseline(char *line, struct rule *r, int lineno)
   int rxflags = REG_EXTENDED;
   static int nrule;
 
-  memset(r, 0, sizeof r);
+  memset(r, 0, sizeof *r);
   r->nrule = nrule;
 
   if ( !readescstring(buffer, &line) )
